"""Provides various field types."""
import re
import datetime
from bson import ObjectId
from pymongo.errors import InvalidId
from tavi import EmbeddedList
from tavi.base.fields import BaseField
from tavi.documents import EmbeddedDocument
from tavi.errors import TaviTypeError

class BooleanField(BaseField):
    """Represents a boolean field for a Mongo Document. Supports all the
    validations of *BaseField*.

    """
    def validate(self, instance, value):
        super(BooleanField, self).validate(instance, value)

        if not isinstance(value, bool):
            instance.errors.add(self.name, "must be a valid boolean")

class DateTimeField(BaseField):
    """Represents a naive datetime for a Mongo Document.
    Supports all the validations in *BaseField*.

    """
    def validate(self, instance, value):
        """Validates the field."""
        super(DateTimeField, self).validate(instance, value)

        if None != value and not isinstance(value, datetime.datetime):
            instance.errors.add(self.name, "must be a valid date and time")

class FloatField(BaseField):
    """Represents a floating point number for a Mongo Document.

    Supports all the validations in *BaseField* and the following:

    min_value -- validates the minimum value the field value can be
    max_value -- validates the maximum value the field value can be

    """
    def __init__(self, name, min_value=None, max_value=None, **kwargs):
        super(FloatField, self).__init__(name, **kwargs)

        self.min_value = None if None == min_value else float(min_value)
        self.max_value = None if None == max_value else float(max_value)

    def validate(self, instance, value):
        """Validates the field."""
        super(FloatField, self).validate(instance, value)

        if isinstance(value, int):
            value = float(value)

        if not isinstance(value, float):
            instance.errors.add(self.name, "must be a float")

        if not None == self.min_value and value < self.min_value:
            instance.errors.add(self.name,
                "is too small (minimum is %s)" % self.min_value)

        if not None == self.max_value and value > self.max_value:
            instance.errors.add(self.name,
                "is too big (maximum is %s)" % self.max_value)

class IntegerField(BaseField):
    """Represents a integer number for a Mongo Document.

    Supports all the validations in *BaseField* and the following:

    min_value -- validates the minimum value the field value can be
    max_value -- validates the maximum value the field value can be

    """
    def __init__(self, name, min_value=None, max_value=None, **kwargs):
        super(IntegerField, self).__init__(name, **kwargs)

        self.min_value = None if None == min_value else int(min_value)
        self.max_value = None if None == max_value else int(max_value)

    def validate(self, instance, value):
        """Validates the field."""
        super(IntegerField, self).validate(instance, value)

        if not isinstance(value, int):
            instance.errors.add(self.name, "must be a integer")

        if not None == self.min_value and value < self.min_value:
            instance.errors.add(self.name,
                "is too small (minimum is %s)" % self.min_value)

        if not None == self.max_value and value > self.max_value:
            instance.errors.add(self.name,
                "is too big (maximum is %s)" % self.max_value)

class ObjectIdField(BaseField):
    """Represents an Object Id generated by Mongo. Supports all the validations
    in *BaseField*.

    """
    def __set__(self, instance, raw_value):
        if raw_value is not None:
            try:
                value = ObjectId(raw_value)
            except InvalidId:
                value = raw_value
        else:
            value = None

        super(ObjectIdField, self).__set__(instance, value)

    def validate(self, instance, value):
        """Validates the field."""
        super(ObjectIdField, self).validate(instance, value)

        if value is not None:
            if not isinstance(value, ObjectId):
                instance.errors.add(self.name, "must be a valid Object Id")

class StringField(BaseField):
    """Represents a String field for a Mongo Document.

    Supports all the validations in *BaseField* and the following:

    length     -- validates the field value has an exact length; default is
                  *None*

    min_length -- ensures field has a minimum number of characters; default is
                  *None*

    max_length -- ensures field is not more than a maximum number of
                  characters; default is *None*

    pattern    -- validates the field matches the given regular expression
                  pattern; default is *None*

    """
    def __init__(self, name, length=None, min_length=None, max_length=None,
            pattern=None, **kwargs):
        super(StringField, self).__init__(name, **kwargs)

        self.length     = length
        self.min_length = min_length
        self.max_length = max_length
        self.regex      = re.compile(pattern) if pattern else None

    def __set__(self, instance, unstripped_value):
        value = unstripped_value.strip() if unstripped_value else None
        super(StringField, self).__set__(instance, value)

    def validate(self, instance, value):
        """Validates the field."""
        super(StringField, self).validate(instance, value)
        val_length = len(value) if value else None

        if self.required and '' == value:
            instance.errors.add(self.name, "is required")

        if self.length and self.length != val_length:
            instance.errors.add(self.name,
                "is the wrong length (should be %s characters)" % self.length)

        if self.min_length and val_length < self.min_length:
            instance.errors.add(self.name,
                "is too short (minimum is %s characters)" % self.min_length)

        if self.max_length and val_length > self.max_length:
            instance.errors.add(self.name,
                "is too long (maximum is %s characters)" % self.max_length)

        if self.regex and self.regex.match(value) is None:
            instance.errors.add(self.name, "is in the wrong format")

class EmbeddedField(BaseField):
    """Represents an embedded Mongo document. Raises a TaviTypeError if *doc*
    is not a tavi.document.EmbeddedDocument.

    """
    def __init__(self, name, doc, **kwargs):
        super(EmbeddedField, self).__init__(name, **kwargs)
        doc_instance = doc()

        if not isinstance(doc_instance, EmbeddedDocument):
            raise TaviTypeError(
                    "expected %s to be a subclass of "
                    "tavi.document.EmbeddedDocument" %
                    doc_instance.__class__
                    )

        self.doc_class = doc
        self.value = self.default or doc_instance

    def __get__(self, instance, owner):
        return self.value

    def __set__(self, instance, value):
        if value:
            if not self.value: self.value = self.doc_class()
            for field in value.fields:
                embedded_value = getattr(value, field, None)
                setattr(self.value, field, embedded_value)
        else:
            self.value = value


class ListField(BaseField):
    """Represents a list of embedded document fields."""
    def __get__(self, instance, owner):
        if self.attribute_name not in instance.__dict__:
            setattr(instance, self.attribute_name, EmbeddedList(self.name))
        return getattr(instance, self.attribute_name)

    def __set__(self, instance, value):
        pass



"""Provides base document support."""
import collections
from bson.json_util import dumps, loads
from tavi.errors import Errors
from tavi.base.fields import BaseField
import tavi

def get_field_attr(cls, field):
    """Custom function for retrieving a tavi.field attribute. Handles nested
    attributes for embedded documents as well as embedded lists of documents.

    """
    value = getattr(cls, field)
    if isinstance(value, BaseDocument):
        return value.field_values
    if isinstance(value, collections.MutableSequence):
        return [v.field_values for v in value]
    else:
        return value

def set_field_attr(cls, field, value):
    """Custom function for setting a tavi.field attribute."""
    field_descriptor = cls._field_descriptors[field]
    if not value: value = field_descriptor.default
    if isinstance(field_descriptor , tavi.fields.EmbeddedField) and not value: value = {}
    if isinstance(value, dict): value = field_descriptor.doc_class(**value)
    setattr(cls, field, value)

class BaseDocumentMetaClass(type):
    """MetaClass for BaseDocuments. Handles initializing the list of fields for
    the BaseDocument.

    """
    def __init__(cls, name, bases, attrs):
        super(BaseDocumentMetaClass, cls).__init__(name, bases, attrs)

        sorted_fields = sorted(
            [field for field in attrs.iteritems() if isinstance(field[1],
                BaseField) and field[1].persist],
            key = lambda i: i[1].creation_order
        )

        cls._field_descriptors = collections.OrderedDict(sorted_fields)

class BaseDocument(object):
    """Base class for Mongo Documents. Provides basic field support."""
    __metaclass__ = BaseDocumentMetaClass

    def __init__(self, **kwargs):
        self._errors = Errors()
        for field in self.fields:
            field_descriptor = self._field_descriptors[field]
            set_field_attr(self, field, kwargs.get(field))

            field_value = getattr(self, field)
            if hasattr(field_value, "owner"):
                field_value.owner = self

    @property
    def field_values(self):
        """Returns a dictionary containing all fields and their values."""
        return { field: get_field_attr(self, field) for field in self.fields }

    @property
    def errors(self):
        """Returns a tavi.Errors object that contains any errors for the
        Document.

        """
        return self._errors

    @property
    def fields(self):
        """Returns the list of fields for the Document."""
        return self._field_descriptors.keys()

    @property
    def valid(self):
        """Indicates if all the fields in the Document are valid."""
        return 0 == self.errors.count

    def to_json(self, fields=None):
        """Convert Document model object to JSON. Optionally, specify which
        fields should be serialized.

        """
        if fields:
            return dumps({ field: self.field_values[field] for field in fields })
        else:
            return dumps(self.field_values)

    @classmethod
    def from_json(cls, json_str):
        """Deserialize a JSON string into a Document model object."""
        attrs = loads(json_str)
        return cls(**attrs)



"""Support for timing blocks of code."""
from __future__ import with_statement
import time

class Timer(object):
    """An object use to time a block of code."""
    def __init__(self):
        self._start = None
        self._finish = None

    def __enter__(self):
        self._start = time.time()

    def __exit__(self, type_, value, traceback):
        self._finish = time.time()

    def duration_in_seconds(self):
        """The amount of time taken to execute block of code. Rounded to
        nearest millisecond.

        """
        return round(self._finish - self._start, 3)



"""Defines custom errors."""
import inflection
from tavi.utils import flatten

class TaviError(Exception):
    """Base error for all Tavi errors."""
    pass

class TaviTypeError(TaviError):
    """Raised when an operation or function is applied to an object of
    inappropriate type.

    """
    pass

class TaviConnectionError(TaviError):
    """Raised when Tavi cannot connect to Mongo."""
    pass

class Errors(object):
    """Provides a dictionary-like object that is used for handing error messages
    for fields.

    """

    def __init__(self):
        self._errors = {}

    @property
    def count(self):
        """Returns the number of error messages."""
        return len(flatten(self._errors.values()))

    @property
    def full_messages(self):
        """Returns all the full error messages as a list."""
        return flatten([self.full_messages_for(field) for field in
            self._errors])

    def add(self, field, message):
        """Adds *message* to the error messages on *field*. More than one error
        can be added to the same *field*.

        """
        if field not in self._errors:
            self._errors[field] = []
        self._errors[field].append(message)

    def clear(self, field):
        """Clear the error messages."""
        self._errors[field] = []

    def full_messages_for(self, field):
        """Returns all the full error messages for a given *field* as a list."""
        humanized_field = inflection.titleize(inflection.humanize(field))
        return ["%s %s" % (humanized_field, msg) for msg in self.get(field)]

    def get(self, field):
        """Return error messages for *field*."""
        return self._errors[field]



# I prefer Markdown to reStructuredText. PyPi does not. This allows people to
# install and not get any errors.
try:
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst')
except (IOError, ImportError):
    long_description = (
        "Tavi (as in `Rikki Tikki Tavi "
        "<http://en.wikipedia.org/wiki/Rikki-Tikki-Tavi>`_) "
        "is an extremely thin Mongo object mapper for Python. It is a thin "
        "abstraction over `pymongo <http://api.mongodb.org/python/current/>`_ "
        "that allows you to easily model your applications and persist your "
        "data in MongoDB. See `README.md <http://github.com/bnadlerjr/tavi>`_ "
        "for more details."
    )

from setuptools import setup

setup(
    name='Tavi',
    version='0.0.7',
    author='Bob Nadler Jr.',
    author_email='bnadlerjr@gmail.com',
    packages=['tavi', 'tavi.base', 'tavi.utils'],
    url='https://github.com/bnadlerjr/tavi',
    license='LICENSE.txt',
    description='Super thin Mongo object mapper for Python.',
    long_description=long_description,
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        "inflection >= 0.2.0",
        "pymongo >= 2.5.2"
    ]
)



"""Various utility functions."""
def flatten(target):
    """Flatten a list with nested lists into a single list.

    For example::
        flatten([1, [2, 3]) # => [1, 2, 3]

    """
    return [item for sublist in target for item in sublist]
